from __future__ import annotations

import inspect
import os
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, NamedTuple, TypeVar
from urllib.parse import urlparse

import anyio

from eventiq.exceptions import BrokerConnectionError

from .imports import import_from_string
from .logging import LoggerMixin
from .settings import BrokerSettings, UrlBrokerSettings
from .types import DecodedMessage, DefaultAction, Message, Timeout
from .utils import format_topic, to_float

if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectSendStream
    from pydantic import AnyUrl

    from eventiq import Consumer


R = TypeVar("R", bound=Any)


class BulkMessage(NamedTuple):
    topic: str
    body: bytes
    headers: dict[str, str]
    kwargs: dict[str, Any]


class Broker(Generic[Message, R], LoggerMixin, ABC):
    """Base broker class
    :param description: Broker (Server) Description
    :param encoder: Encoder (Serializer) class
    :param decoder: Decoder (Deserializer) class.
    """

    protocol: str
    protocol_version: str = ""

    Settings: type[BrokerSettings] = BrokerSettings

    WILDCARD_ONE: str
    WILDCARD_MANY: str

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if not inspect.isabstract(cls):
            protocol = getattr(cls, "protocol", None)
            if protocol is None:
                msg = f"Broker subclass {cls} must define a protocol"
                raise ValueError(msg)

    def __init__(
        self,
        *,
        name: str = "default",
        description: str | None = None,
        default_on_exc: DefaultAction = "nack",
        default_consumer_timeout: Timeout = 300,
        tags: list[str] | None = None,
        asyncapi_extra: dict[str, Any] | None = None,
        validate_error_delay: int = 3600 * 12,
    ) -> None:
        self.name = name
        self.description = description or type(self).__name__
        self.default_on_exc = default_on_exc
        self.tags = tags
        self.async_api_extra = asyncapi_extra or {}
        self.default_consumer_timeout = to_float(default_consumer_timeout)
        self.validate_error_delay = validate_error_delay

    def __repr__(self) -> str:
        return type(self).__name__

    def should_nack(self, raw_message: Message) -> bool:
        return False

    def format_topic(self, topic: str) -> str:
        return format_topic(topic, self.WILDCARD_ONE, self.WILDCARD_MANY)

    @abstractmethod
    def get_info(self) -> dict[str, Any]:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def decode_message(raw_message: Message) -> DecodedMessage:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """Return broker connection status."""
        raise NotImplementedError

    @abstractmethod
    async def publish(
        self,
        topic: str,
        body: bytes,
        *,
        headers: dict[str, str],
        **kwargs: Any,
    ) -> R:
        raise NotImplementedError

    async def bulk_publish(
        self,
        messages: list[BulkMessage],
        topic: str | None = None,
    ) -> None:
        """
        Default implementation of bulk publish uses task groups to publish messages concurrently.
        """
        async with anyio.create_task_group() as tg:
            for msg_topic, body, headers, kwargs in messages:
                message_topic = topic or msg_topic
                tg.start_soon(self._publish_task, message_topic, body, headers, kwargs)

    async def _publish_task(
        self, topic: str, body: bytes, headers: dict[str, str], kwargs: dict[str, Any]
    ) -> None:
        await self.publish(topic, body, headers=headers, **kwargs)

    @abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def disconnect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def ack(self, raw_message: Message) -> None:
        raise NotImplementedError

    @abstractmethod
    async def nack(self, raw_message: Message, delay: int | None = None) -> None:
        raise NotImplementedError

    @abstractmethod
    async def sender(
        self,
        group: str,
        consumer: Consumer,
        send_stream: MemoryObjectSendStream[Message],
    ) -> None:
        raise NotImplementedError

    def get_num_delivered(self, raw_message: Message) -> int | None:
        return getattr(raw_message, "num_delivered", None)

    @classmethod
    def from_settings(cls, settings: BrokerSettings, **kwargs: Any) -> Broker:
        return cls(**settings.model_dump(), **kwargs)

    @classmethod
    def _from_env(cls, **kwargs: Any) -> Broker:
        return cls.from_settings(cls.Settings(**kwargs))

    @classmethod
    def from_env(
        cls,
        **kwargs: Any,
    ) -> Broker:
        if cls == Broker:
            type_name = os.getenv("BROKER_CLASS", "eventiq.backends.stub:StubBroker")
            cls = import_from_string(type_name)
        return cls._from_env(**kwargs)


class UrlBroker(Broker[Message, R], ABC):
    settings: type[UrlBrokerSettings]
    error_msg = "Broker not connected"

    def __init__(
        self,
        *,
        url: AnyUrl,
        connection_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.url = str(url)
        self.connection_options = connection_options or {}

    @property
    def connection_error(self) -> BrokerConnectionError:
        return BrokerConnectionError(self.error_msg)

    def get_info(self) -> dict[str, Any]:
        parsed = urlparse(self.url)
        return {
            "host": parsed.hostname,
            "pathname": parsed.path,
        }



from __future__ import annotations

from typing import Any

from .exceptions import DecodeError
from .types import AnyType, Decoder, RawData, T


class JsonDecoder:
    CONTENT_TYPE = "application/json"

    def __init__(self, strict: bool = False, context: Any = None) -> None:
        self.strict = strict
        self.context = context

    def decode(self, data: RawData, as_type: type[T] | None = None) -> T | Any:
        try:
            if as_type:
                return as_type.model_validate_json(
                    data,
                    strict=self.strict,
                    context=self.context,
                )
            return AnyType.validate_json(data, strict=self.strict, context=self.context)
        except Exception as e:
            raise DecodeError from e


DEFAULT_DECODER: Decoder = JsonDecoder()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Callable

from eventiq.middleware import CloudEventType, Middleware
from eventiq.utils import to_async

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from eventiq import Consumer, Service


class ErrorHandlerMiddleware(Middleware[CloudEventType]):
    def __init__(
        self,
        service: Service,
        callback: Callable[
            [Service, Consumer, CloudEventType, Exception | None], Awaitable[Any]
        ],
        errors: type[Exception] | tuple[type[Exception]] = Exception,
    ) -> None:
        super().__init__(service)
        if not asyncio.iscoroutinefunction(callback):
            callback = to_async(callback)
        self.callback = callback
        self.exc = errors

    async def after_process_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        if exc and isinstance(exc, self.exc):
            await self.callback(self.service, consumer, message, exc)



import asyncio
from collections.abc import AsyncGenerator, AsyncIterator
from contextlib import asynccontextmanager, suppress
from unittest.mock import AsyncMock, MagicMock

import pytest
import pytest_asyncio

from eventiq import CloudEvent, Consumer, GenericConsumer, Service
from eventiq.backends.stub import StubBroker
from eventiq.consumer import FnConsumer
from eventiq.middleware import Middleware
from eventiq.middlewares.dlx import DeadLetterQueueMiddleware
from eventiq.middlewares.error import ErrorHandlerMiddleware
from eventiq.middlewares.healthcheck import HealthCheckMiddleware
from eventiq.middlewares.perf_counter import PerfCounterMiddleware
from eventiq.middlewares.retries import RetryMiddleware
from eventiq.utils import utc_now


@pytest_asyncio.fixture(scope="session")
def event_loop():
    loop = asyncio.new_event_loop()
    yield loop
    loop.stop()


@pytest.fixture(scope="session")
def middleware():
    class EmptyMiddleware(Middleware):
        pass

    return EmptyMiddleware


@pytest.fixture()
def broker():
    return StubBroker()


@pytest.fixture()
def service(broker, middleware):
    svc = Service(name="test_service", broker=broker, state={int: 5})
    svc.add_middleware(middleware)
    svc.add_middleware(DeadLetterQueueMiddleware)
    svc.add_middleware(RetryMiddleware)
    svc.add_middleware(HealthCheckMiddleware)
    svc.add_middleware(ErrorHandlerMiddleware, callback=AsyncMock())
    svc.add_middleware(PerfCounterMiddleware)
    return svc


@pytest.fixture(scope="session")
def handler():
    async def example_handler(message: CloudEvent, x: int) -> int:
        assert isinstance(message, CloudEvent)
        assert x == 5
        return 42

    return example_handler


@pytest.fixture()
def test_consumer(service, handler):
    service.subscribe(handler, topic="test_topic", name="test_consumer")
    return service.consumer_group.consumers["test_consumer"]


@pytest.fixture()
def generic_test_consumer(service) -> Consumer:
    generic_consumer_name = "test_generic_consumer"

    @service.subscribe(topic="test_topic")
    class TestConsumer(GenericConsumer[CloudEvent]):
        name = generic_consumer_name

        async def process(self, message: CloudEvent, x: int = 3) -> int:
            assert isinstance(message, CloudEvent)
            assert x == 5
            return 42

    return service.consumer_group.consumers[generic_consumer_name]


@pytest.fixture()
def ce(service) -> CloudEvent:
    ce_ = CloudEvent.new(
        {"today": utc_now().date().isoformat(), "arr": [1, "2", 3.0]},
        type="TestEvent",
        topic="test_topic",
    )
    ce_.set_context(service, None, {})
    return ce_


@pytest.fixture()
def mock_consumer():
    # this is workaround for inspect.getsignature() of AsyncMock
    # https://github.com/python/cpython/issues/96127
    mock = MagicMock(return_value=AsyncMock())
    mock.__annotations__ = {"message": CloudEvent, "return": None}
    mock.__name__ = "mock_consumer"
    return mock


@asynccontextmanager
async def service_context(service) -> AsyncIterator[None]:
    task = asyncio.create_task(service.run(enable_signal_handler=False))
    await asyncio.sleep(0)
    yield
    with suppress(asyncio.CancelledError):
        task.cancel()
        await task


@pytest_asyncio.fixture()
async def running_service(service: Service, mock_consumer) -> AsyncGenerator:
    consumer: Consumer = FnConsumer(
        fn=mock_consumer, event_type=CloudEvent, topic="test_topic"
    )
    service.consumer_group.add_consumer(consumer)

    async with service_context(service):
        yield service



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Mapping
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Callable, Generic, NamedTuple

from typing_extensions import ParamSpec

from eventiq.exceptions import Fail, Retry, Skip
from eventiq.logging import LoggerMixin
from eventiq.middleware import CloudEventType, Middleware

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service


P = ParamSpec("P")

DelayGenerator = Callable[[CloudEventType, Exception], int]


class MessageStatus(NamedTuple):
    status: str
    delay: int | None


def expo(factor: int = 1) -> DelayGenerator:
    def _expo(message: CloudEvent, _: Exception) -> int:
        return factor * message.age.seconds

    return _expo


def constant(interval: int = 30) -> DelayGenerator:
    def _constant(*_: Any) -> int:
        return interval

    return _constant


class AbstractRetryStrategy(ABC, Generic[CloudEventType]):
    @abstractmethod
    def maybe_retry(
        self,
        *,
        service: Service,
        message: CloudEventType,
        exc: Exception,
    ) -> None:
        raise NotImplementedError


class RetryStrategy(
    AbstractRetryStrategy[CloudEventType], Generic[P, CloudEventType], LoggerMixin
):
    def __init__(
        self,
        throws: tuple[type[Exception], ...] = (),
        delay_generator: Callable[P, DelayGenerator] | None = None,
        min_delay: int = 2,
        log_exceptions: bool = True,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> None:
        if Fail not in throws:
            throws = (*throws, Fail)
        self.throws = throws
        self.min_delay = min_delay
        self.log_exceptions = log_exceptions
        self.delay_generator = (
            delay_generator(*args, **kwargs) if delay_generator else expo()
        )

    def retry(self, message: CloudEventType, exc: Exception) -> None:
        delay = getattr(exc, "delay", None)
        if delay is None:
            delay = self.delay_generator(message, exc)
        delay = max(delay, self.min_delay)
        if self.log_exceptions:
            self.logger.warning(
                "Will retry message %s in %d seconds.",
                message.id,
                delay,
                exc_info=exc,
            )
        raise Retry(delay=delay) from exc

    def fail(self, message: CloudEventType, exc: Exception) -> None:
        self.logger.exception(
            "Retry limit exceeded for message %s",
            message.id,
            exc_info=exc,
        )
        raise Fail(reason="Retry limit exceeded") from exc

    def maybe_retry(
        self,
        *,
        service: Service,
        message: CloudEventType,
        exc: Exception,
    ) -> None:
        if not (self.throws and isinstance(exc, self.throws)):
            self.retry(message, exc)
        else:
            self.fail(message, exc)


class MaxAge(RetryStrategy[P, CloudEventType]):
    def __init__(
        self,
        max_age: timedelta | dict[str, Any] = timedelta(seconds=60),
        **extra: Any,
    ) -> None:
        super().__init__(**extra)
        if isinstance(max_age, Mapping):
            max_age = timedelta(**max_age)
        self.max_age: timedelta = max_age

    def maybe_retry(
        self,
        *,
        service: Service,
        message: CloudEventType,
        exc: Exception,
    ) -> None:
        if message.age <= self.max_age:
            super().maybe_retry(service=service, message=message, exc=exc)
        else:
            self.fail(message, exc)


class MaxRetries(RetryStrategy[P, CloudEventType]):
    def __init__(self, max_retries: int = 3, **extra: Any) -> None:
        super().__init__(**extra)
        self.max_retries = max_retries

    def maybe_retry(
        self,
        *,
        service: Service,
        message: CloudEventType,
        exc: Exception,
    ) -> None:
        retries = service.broker.get_num_delivered(message.raw)
        if retries is None:
            self.logger.warning(
                "Retries property not found in message, backing off to message.age.seconds",
            )
            retries = int(message.age.total_seconds() ** 0.5)
        if retries <= self.max_retries:
            super().maybe_retry(service=service, message=message, exc=exc)
        else:
            self.fail(message, exc)


class RetryWhen(RetryStrategy[P, CloudEventType]):
    def __init__(
        self,
        retry_when: Callable[[CloudEventType, Exception], bool],
        **extra: Any,
    ) -> None:
        super().__init__(**extra)
        self.retry_when = retry_when

    def maybe_retry(
        self,
        *,
        service: Service,
        message: CloudEventType,
        exc: Exception,
    ) -> None:
        if self.retry_when(message, exc):
            super().maybe_retry(service=service, message=message, exc=exc)
        else:
            self.fail(message, exc)


class RetryMiddleware(Middleware[CloudEventType]):
    """Retry Message Middleware.
    Supported retry strategies:
    - `MaxAge` (default) - retry with exponential backoff up to max_age
    - `MaxRetries` - retry up to N times (currently supported only by nats)
    - `RetryWhen` - provide custom callable to determine weather message should be retried.
    """

    def __init__(
        self,
        service: Service,
        default_retry_strategy: RetryStrategy = MaxAge(max_age=timedelta(hours=6)),
    ) -> None:
        super().__init__(service)
        self.default_retry_strategy = default_retry_strategy

    async def after_process_message(
        self,
        *,
        consumer: Consumer,
        message: CloudEventType,
        exc: Exception | None = None,
        **_: Any,
    ) -> None:
        if exc is None or isinstance(exc, (Retry, Fail, Skip)):
            return

        retry_strategy = consumer.options.get(
            "retry_strategy",
            self.default_retry_strategy,
        )
        retry_strategy.maybe_retry(service=self.service, message=message, exc=exc)



from eventiq.backends.nats import NatsBroker


def test_nats_broker_topic_pattern():
    broker = NatsBroker(url="nats://localhost:4222")
    pattern = "events.{param}.*"
    result = broker.format_topic(pattern)
    assert result == "events.*.>"



from __future__ import annotations

import functools
from inspect import Parameter, signature
from typing import TYPE_CHECKING, Any, Callable

from typing_extensions import Concatenate, ParamSpec

from .exceptions import DependencyError

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from .types import CloudEventType

P = ParamSpec("P")


def resolved_func(
    func: Callable[Concatenate[CloudEventType, P], Awaitable[Any]],
) -> Callable[Concatenate[CloudEventType, P], Awaitable[Any]]:
    sig = signature(func)
    params = {
        k: (v.annotation, v.default)
        for k, v in sig.parameters.items()
        if k not in {"message", "args", "kwargs"}
        and v.kind
        not in {
            Parameter.POSITIONAL_ONLY,
            Parameter.VAR_POSITIONAL,
            Parameter.VAR_KEYWORD,
        }
    }

    if not params:
        return func

    @functools.wraps(func)
    async def wrapped(
        message: CloudEventType, *args: P.args, **kwargs: P.kwargs
    ) -> Any:
        state = message.service.state

        for k, v in params.items():
            annotation, default = v
            if annotation in state:
                kwargs[k] = state[annotation]
            elif default is Parameter.empty:
                err = f"Missing dependency {k}: {annotation}"
                raise DependencyError(err)

        return await func(message, *args, **kwargs)

    return wrapped



from pydantic import BaseModel

from eventiq import CloudEvent, Publishes, Service
from eventiq.backends.nats import JetStreamBroker
from eventiq.models import Command

broker = JetStreamBroker(url="nats://nats:password@localhost:4222")


class MyData(BaseModel):
    """Main data for service"""

    counter: int
    info: str


class MyEvent(CloudEvent[MyData]):
    """Some custom event"""


class MyCommand(Command[int], topic="commands.run"):
    """Command representing current number of items"""


service = Service(
    name="example-service",
    version="1.0",
    broker=broker,
    publishes=[
        Publishes(
            type=MyEvent,
            topic="test.topic.{param}.*",
            tags=["tag2"],
            summary="Publishes when X happens",
            parameters={
                "param": {
                    "enum": ["x", "y", "z"],
                    "description": "Description for param",
                },
            },
        ),
    ],
    tags_metadata=[{"name": "tag1", "description": "Some tag 1"}],
)


@service.subscribe(
    topic="test.topic.{param}.*",
    tags=["tag1"],
    parameters={
        "param": {
            "enum": ["a", "b", "c"],
            "description": "Some description",
            "examples": ["a", "b"],
        },
    },
)
async def example_handler(message: MyEvent):
    """Consumer for processing MyEvent(s)"""
    print(f"Received Message {message.id} with data: {message.data}")


@service.subscribe
async def example_run(message: MyCommand):
    """Consumer for processing MyCommands(s)"""
    print(f"Received Message {message.id} with data: {message.data}")



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Any, TypedDict, TypeVar

from pydantic import AnyUrl, UrlConstraints
from redis.asyncio import Redis

from eventiq.broker import UrlBroker, UrlBrokerSettings
from eventiq.results import ResultBackend

if TYPE_CHECKING:
    from anyio.streams.memory import MemoryObjectSendStream

    from eventiq import Consumer
    from eventiq.types import DecodedMessage

RedisUrl = Annotated[AnyUrl, UrlConstraints(allowed_schemes=["redis", "rediss"])]


class RMessage(TypedDict):
    type: bytes | str
    pattern: bytes | str | None
    channel: bytes | str
    data: bytes


RedisRawMessage = TypeVar("RedisRawMessage", bound=RMessage)


class RedisBroker(
    UrlBroker[RedisRawMessage, None], ResultBackend[RedisRawMessage, None]
):
    """
    Broker implementation based on redis PUB/SUB and aioredis package
    :param kwargs: base class arguments
    """

    Settings = UrlBrokerSettings[RedisUrl]
    protocol = "redis"

    WILDCARD_ONE = "*"
    WILDCARD_MANY = "*"

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._redis: Redis | None = None

    @staticmethod
    def decode_message(raw_message: RedisRawMessage) -> DecodedMessage:
        return raw_message["data"], None

    @property
    def is_connected(self) -> bool:
        if self.redis.connection:
            return self.redis.connection.is_connected
        return False

    @property
    def redis(self) -> Redis:
        if self._redis is None:
            raise self.connection_error
        return self._redis

    async def sender(
        self, group: str, consumer: Consumer, send_stream: MemoryObjectSendStream
    ) -> None:
        async with self.redis.pubsub() as sub:
            await sub.psubscribe(consumer.topic)
            async with send_stream:
                while True:
                    message = await sub.get_message(ignore_subscribe_messages=True)
                    if message:
                        await send_stream.send(message)

    async def disconnect(self) -> None:
        await self.redis.close()

    async def connect(self) -> None:
        self._redis = Redis.from_url(self.url, **self.connection_options)

    async def publish(
        self,
        topic: str,
        body: bytes,
        **kwargs: Any,
    ) -> None:
        await self.redis.publish(topic, body)

    async def store_result(self, key: str, result: bytes) -> None:
        await self.redis.set(key, result)

    async def get_result(self, key: str) -> Any | None:
        return await self.redis.get(key)

    async def ack(self, raw_message: RedisRawMessage) -> None:
        pass

    async def nack(
        self, raw_message: RedisRawMessage, delay: int | None = None
    ) -> None:
        await self.redis.publish(raw_message["channel"], raw_message["data"])



from eventiq import CloudEvent, GenericConsumer, Service
from eventiq.backends.stub import StubBroker

broker = StubBroker()

service = Service(name="example-service", broker=broker)


@service.subscribe(topic="example.topic")
class MyConsumer(GenericConsumer[CloudEvent]):
    # optionally replace `CloudEvent` with more specific class
    name = "example_consumer"

    async def process(self, message: CloudEvent):
        print(f"Received Message {message.id} with data: {message.data}")
        await self.publish(
            message.copy(update={"topic": "example.topic2", "data": "new data"})
        )



from __future__ import annotations

import functools
import json
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Literal

from pydantic.alias_generators import to_camel
from pydantic.json_schema import models_json_schema
from pydantic_asyncapi.common import Reference, Tag
from pydantic_asyncapi.v3 import (
    AsyncAPI,
    Channel,
    Components,
    Info,
    Message,
    Operation,
    Parameter,
    Server,
)

from eventiq.utils import TOPIC_PATTERN

TOPIC_TRANSLATION = str.maketrans({"{": "", "}": "", ".": "_", "*": "all"})
PREFIX = "#/components/schemas/"

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from pydantic import BaseModel

    from eventiq import Broker, CloudEvent, Service
    from eventiq.consumer import Consumer
    from eventiq.models import Publishes
    from eventiq.types import Parameter as ParamDict


def save_async_api_to_file(spec: BaseModel, path: Path, fmt: str) -> None:
    data = spec.model_dump(by_alias=True, exclude_none=True, exclude_unset=True)

    with open(path, "w") as f:
        if fmt == "yaml":
            import yaml

            yaml.dump(data, f)
        else:
            json.dump(data, f)


def get_all_models_schema(service: Service) -> Any:
    all_models: list[tuple[type[CloudEvent], Literal["validation"]]] = [
        (m.type, "validation") for m in service.publishes
    ]
    for consumer in service.consumers.values():
        all_models.extend([(m.type, "validation") for m in consumer.publishes])
        all_models.append((consumer.event_type, "validation"))

    _, top_level_schema = models_json_schema(
        all_models,
        ref_template=f"{PREFIX}{{model}}",
    )
    return top_level_schema.get("$defs", {})


def snake_case_to_title(name: str) -> str:
    return name.replace("_", " ").title()


def get_tag_list(tags: dict[str, Tag], taggable: Iterable | None) -> list[Tag]:
    tag_list = []
    for t in taggable or []:
        if t not in tags:
            tags[t] = Tag(name=t)
        tag_list.append(tags[t])
    return tag_list


def generate_channel_id(topic: str) -> str:
    topic_snake = topic.translate(TOPIC_TRANSLATION)
    return to_camel(topic_snake)


def get_topic_parameters(
    topic: str,
    parameters: dict[str, ParamDict],
) -> dict[str, Parameter]:
    result_params = {}
    for k in topic.split("."):
        if TOPIC_PATTERN.fullmatch(k):
            param_name = k[1:-1]
            param = parameters.get(param_name, {"description": param_name})
            result_params[param_name] = Parameter(**param)
    return result_params


def generate_receive_operation(
    consumer: Consumer,
    service: Service,
    channels_params: dict[str, Any],
    spec: AsyncAPI,
    tags: dict[str, Tag],
) -> None:
    event_type: str = consumer.event_type.__name__
    channel_id = generate_channel_id(consumer.topic)
    params = get_topic_parameters(consumer.topic, consumer.parameters)
    for k, v in params.items():
        channels_params[channel_id].setdefault(k, v)
    content_type = (
        consumer.decoder.CONTENT_TYPE
        if consumer.decoder
        else service.decoder.CONTENT_TYPE
    )
    message = Message(
        name=event_type,
        title=event_type,
        description=consumer.event_type.__doc__,
        contentType=content_type,
        payload=Reference(ref=f"#/components/schemas/{event_type}"),
        **consumer.asyncapi_extra.get("message", {}),
    )
    if spec.components is None:
        spec.components = Components()
    if spec.components.messages is None:
        spec.components.messages = {}
    spec.components.messages[event_type] = message

    channel = Channel(
        address=consumer.topic,
        servers=[Reference(ref=f"#/servers/{service.broker.name}")],
        messages={
            event_type: Reference(ref=f"#/channels/{channel_id}/messages/{event_type}"),
        },
        parameters=channels_params[channel_id],
        tags=get_tag_list(tags, consumer.tags),
        **consumer.asyncapi_extra.get("channel", {}),
    )
    if spec.channels is None:
        spec.channels = {}
    spec.channels[channel_id] = channel

    operation_id = f"{to_camel(consumer.name)}Receive"
    operation = Operation(
        action="receive",
        title=f"{snake_case_to_title(consumer.name)} Receive",
        summary=consumer.description,
        channel=Reference(ref=f"#/channels/{channel_id}"),
        **consumer.asyncapi_extra.get("operation", {}),
    )
    if spec.operations is None:
        spec.operations = {}
    spec.operations[operation_id] = operation


def generate_send_operation(
    publish_list: list[Publishes],
    broker: Broker,
    spec: AsyncAPI,
    channels_params: dict[str, dict[str, Any]],
    tags: dict[str, Tag],
) -> None:
    for publishes in publish_list:
        event_type = publishes.type.__name__
        params = get_topic_parameters(publishes.topic, publishes.parameters)
        channel_id = generate_channel_id(publishes.topic)
        for k, v in params.items():
            channels_params[channel_id].setdefault(k, v)

        operation_id = f"send{event_type}"
        operation = Operation(
            action="send",
            title=f"Send {event_type}",
            messages=[Reference(ref=f"#/channels/{channel_id}/messages/{event_type}")],
            channel=Reference(ref=f"#/channels/{channel_id}"),
            tags=get_tag_list(tags, publishes.tags),
            summary=publishes.summary,
            **publishes.asyncapi_extra.get("operation", {}),
        )
        if spec.operations is None:
            spec.operations = {}
        spec.operations[operation_id] = operation
        if spec.channels is None:
            spec.channels = {}
        if channel_id not in spec.channels:
            channel = Channel(
                address=publishes.topic,
                servers=[Reference(ref=f"#/servers/{broker.name}")],
                messages={
                    event_type: Reference(ref=f"#/components/messages/{event_type}"),
                },
                parameters=channels_params[channel_id],
                tags=get_tag_list(tags, publishes.tags),
                summary=publishes.summary,
                **publishes.asyncapi_extra.get("channel", {}),
            )
            spec.channels[channel_id] = channel


def populate_spec(service: Service, spec: AsyncAPI) -> None:
    tags = {t["name"]: Tag.model_validate(t) for t in service.tags_metadata}
    channels_params: dict[str, dict[str, Parameter]] = defaultdict(dict)
    for consumer in service.consumers.values():
        generate_send_operation(
            consumer.publishes,
            service.broker,
            spec,
            channels_params,
            tags,
        )
        generate_receive_operation(
            consumer,
            service,
            channels_params,
            spec,
            tags,
        )

    generate_send_operation(
        service.publishes,
        service.broker,
        spec,
        channels_params,
        tags,
    )


@functools.lru_cache
def get_async_api_spec(service: Service) -> AsyncAPI:
    schemas = get_all_models_schema(service)
    spec = AsyncAPI(
        asyncapi="3.0.0",
        info=Info(
            title=service.title,
            version=service.version,
            **service.async_api_extra,
        ),
        defaultContentType=service.encoder.CONTENT_TYPE,
        servers={
            service.broker.name: Server(
                protocol=service.broker.protocol,
                **service.broker.get_info(),
                **service.broker.async_api_extra,
            ),
        },
        components=Components(schemas=schemas),
    )
    populate_spec(service, spec)

    return spec



from .__about__ import __version__
from .broker import Broker
from .consumer import Consumer, ConsumerGroup, GenericConsumer
from .middleware import Middleware
from .models import CloudEvent, Publishes
from .service import Service

__all__ = [
    "__version__",
    "Broker",
    "Consumer",
    "ConsumerGroup",
    "CloudEvent",
    "Publishes",
    "GenericConsumer",
    "Middleware",
    "Service",
]



import logging.config
import sys
from pathlib import Path
from typing import Optional

import anyio
import typer

from .imports import import_from_string
from .logging import get_logger
from .models import CloudEvent
from .service import Service

cli = typer.Typer()

logger = get_logger(__name__, "cli")

if "." not in sys.path:
    sys.path.insert(0, ".")


def import_service(path: str) -> Service:
    service = import_from_string(path)
    if not isinstance(service, Service):
        msg = f"Service must be an instance of Service, got {type(service)}"
        raise TypeError(msg)
    return service


def _build_target_from_opts(
    service: str,
    log_level: Optional[str],
    log_config: Optional[str],
    use_uvloop: Optional[bool],
    debug: Optional[bool],
) -> str:
    cmd = [f"eventiq run {service}"]
    if log_level:
        cmd.append(f"--log-level={log_level}")
    if log_config:
        cmd.append(f"--log-config={log_config}")
    if use_uvloop:
        cmd.append("--use-uvloop")
    if debug:
        cmd.append("--debug")
    return " ".join(cmd)


@cli.command(help="Run service")
def run(
    service: str,
    log_level: Optional[str] = typer.Option(
        None,
        help="Logger level, accepted values are: debug, info, warning, error, critical",
    ),
    log_config: Optional[str] = typer.Option(
        None,
        help="Logging file configuration path.",
    ),
    use_uvloop: Optional[bool] = typer.Option(None, help="Enable uvloop"),
    debug: bool = typer.Option(False, help="Enable debug"),
    reload: Optional[str] = typer.Option(None, help="Hot-reload on provided path"),
) -> None:
    if reload:
        try:
            from watchfiles import run_process
        except ImportError:
            logger.exception(
                "--reload option requires 'watchfiles' installed. Please run 'pip install watchfiles'.",
            )
            return
        logger.info("Watching for changes in: %s", reload)
        target = _build_target_from_opts(
            service,
            log_level,
            log_config,
            use_uvloop,
            debug,
        )
        run_process(
            reload,
            target=target,
            target_type="command",
            callback=logger.info,
            sigint_timeout=30,
            sigkill_timeout=30,
        )

    if log_level:
        logging.basicConfig(level=log_level.upper())
    if log_config:
        logging.config.fileConfig(log_config)
    instance = import_service(service)
    logger.info("Running service: %s", service)
    anyio.run(
        instance.run,
        True,
        backend="asyncio",
        backend_options={"use_uvloop": use_uvloop, "debug": debug},
    )


@cli.command(help="Send message via cli")
def send(
    service: str = typer.Argument(
        ...,
        help="Global service object to import in format {package}.{module}:{service_object}",
    ),
    topic: str = typer.Argument(
        ...,
        help="Topic name",
    ),
    data: str = typer.Argument(
        ...,
        help="Data to send",
    ),
    type: str = typer.Option(
        "CloudEvent",
        help="Message type",
    ),
) -> None:
    svc = import_service(service)

    async def connect_and_send(message: CloudEvent) -> None:
        await svc.broker.connect()
        try:
            await svc.publish(message)
        finally:
            await svc.broker.disconnect()

    message_data = svc.decoder.decode(data)
    message = CloudEvent.new(message_data, type=type, topic=topic)
    anyio.run(connect_and_send, message)


@cli.command(help="Generate AsyncAPI documentation from service")
def docs(
    service: str = typer.Argument(
        ...,
        help="Global service object to import in format {package}.{module}:{service_object}",
    ),
    out: Path = typer.Option("./asyncapi.json", help="Output file path"),
    format: str = typer.Option(
        "json",
        help="Output format. Valid options are 'yaml' and 'json'(default)",
    ),
) -> None:
    from eventiq.asyncapi import get_async_api_spec, save_async_api_to_file

    svc = import_service(service)
    spec = get_async_api_spec(svc)
    save_async_api_to_file(spec, out, format)
    typer.secho(f"Docs saved successfully to {out}", fg="green")
